local function am(msg)
	ChatFrame1:AddMessage(msg);
end

local function getCoinAmount(coinString)
	local amount = 0;
	if not coinString then
		--do nothing
	elseif coinString:find("|T") then
		--coinTextureString
		coinString:gsub("(%d+)\|T([^\|]*)\|t", function(digit, texureStr)
			if texureStr:find("Gold") then
				amount = amount + digit * 10000;
			elseif texureStr:find("Silver") then
				amount = amount + digit * 100;
			elseif texureStr:find("Copper") then
				amount = amount + digit;
			end
		end);
	else
		--coinText
		coinString.gsub("(%d+)%s(%w+)", function(digit, str)
			if str:find("Gold") then
				amount = amount + digit * 10000;
			elseif str:find("Silver") then
				amount = amount + digit * 100;
			elseif str:find("Copper") then
				amount = amount + digit;
			end
		end);
	end
	return amount;
end

local function getShortCoinString(coinString)
	local amount = getCoinAmount(coinString);
	local g = floor(amount / 10000);
	local s = floor((amount % 10000) / 100);
	local b = floor(amount % 100);
	return format("%dg%ds%db", g, s, b);
end

---------------------------------------------------------------------------------------------
local L = {
	[""] = "",
	["QUICKINSPECT"] = "Quick inspect",
	["TARGET"] = "Target",
	["PARTY"] = "Party members",
	["RAID"] = "Raid members",
	["OPENINSPECTOR"] = "Open Inspector",
	["OPTIONS"] = "Options",
	["MERGERECORD"] = "Merge duplicated records",
	["TIMEOUT"] = "Time out (sec)",
	["REFRESH"] = "Refresh",
	["ANNOUNCE"] = "Announce",
	["BOOT"] = "Kick selected",
	["REMOVE"] = "Remove record",
	["A_SAY"] = "Say",
	["A_WHISPER"] = "Whisper ",
	["A_GUILD"] = "Guild",
	["A_PARTY"] = "Party",
	["A_RAID"] = "Raid",
}

local STATE = {
	["INITIAL"] = ORANGE_FONT_COLOR_CODE.."Initialized"..FONT_COLOR_CODE_CLOSE,
	["LOADING"] = ORANGE_FONT_COLOR_CODE.."Inspeting.."..FONT_COLOR_CODE_CLOSE,
	["COMPLETED"] = GREEN_FONT_COLOR_CODE.."Completed"..FONT_COLOR_CODE_CLOSE,
	["DONE"] = GREEN_FONT_COLOR_CODE.."Done"..FONT_COLOR_CODE_CLOSE,
	["INCOMPLETED"] = RED_FONT_COLOR_CODE.."Incomplete"..FONT_COLOR_CODE_CLOSE,
	["OUTOFRANGE"] = YELLOW_FONT_COLOR_CODE.."Out of range"..FONT_COLOR_CODE_CLOSE,
	["TIMEOUT"] = YELLOW_FONT_COLOR_CODE.."TimeOut"..FONT_COLOR_CODE_CLOSE,
	["FAILED"] = YELLOW_FONT_COLOR_CODE.."Failed"..FONT_COLOR_CODE_CLOSE,
}

local HEADERICON = {
	["UP"] = "Interface\\BUTTONS\\Arrow-Up-Up",
	["DOWN"] = "Interface\\BUTTONS\\Arrow-Down-Up",
	["COORD_UP"] = {0, 1.0, 0.375, 1.0},
	["COORD_DOWN"] = {0, 1.0, 0, 0.625},
}

--AchievementDateFont/GameFontNormal/GameFontNormalSmall

local META_STATISTIC = { "name", "state", "info", };
local META_ACHIEVEMENT = { "name", "state", "date", };

local defaultDB = {
	version = "2.0",
	sortType = 0,
	sortSubtype = true,
	mergeRecord = true,
	timeout = 3,
}

local _name, _addon = ...;

local Dewdrop = LibStub("Dewdrop-2.0");
local MessGUI = LibStub("LibMessGUI", true);
local SimpleTimer = LibStub("LibSimpleTimer", true);

local DB;
local InspectorFrame;

local _RecordTable = {};			--	sessionID : {	id, name, flag,
															--								metaData = {	index : fieldName	},
															--								data = {	unitName : unitData = {	fieldName : fieldValue	}	},
															--	}
---------------------------------------------------------------------------------------------inspect core
local FLAG_STATISTIC = 0x00000001;
local COINTEXTURETAG = "|TInterface\\MoneyFrame";

local _ProcessingID;				--current processing session ID
local _ProcessingList = {};	--	index : unitID

local function newUnitData(unitID)
	local unitData = {};
	local name = UnitName(unitID);
	local class = select(2, UnitClassBase(unitID));
	local color = class and RAID_CLASS_COLORS[class] or nil;
	unitData.name = MessGUI:ColorCodeString(name, color);
	return unitData;
end

local function logRecord(id, unitName, ...)
	_RecordTable[id].data[unitName] = _RecordTable[id].data[unitName] or {};
	local unitData = _RecordTable[id].data[unitName];
	for i = 1, select("#", ...), 2 do
		local k, v = select(i, ...);
		unitData[k] = v;
	end
	InspectorFrame:UpdateRecordEntry(id, unitName);
end

local function inspectNext(doRemoval)
	if doRemoval then
		tremove(_ProcessingList, 1);
	end

	local _, unitID = next(_ProcessingList);
	if unitID then
		local name = UnitName(unitID);
		if not name then
			logRecord(_ProcessingID, unitID, "state", STATE["FAILED"]);
		elseif _RecordTable[_ProcessingID].data[name].state == STATE["COMPLETED"] then
			--do nothing
		elseif not UnitIsVisible(unitID) then
			logRecord(_ProcessingID, name, "state", STATE["OUTOFRANGE"]);
		else
			if SetAchievementComparisonUnit(unitID) then
				logRecord(_ProcessingID, name, "state", STATE["LOADING"]);
				SimpleTimer:SetDurationAlarm(_name, DB.timeout, function()
					ClearAchievementComparisonUnit();
					logRecord(_ProcessingID, name, "state", STATE["TIMEOUT"]);
					inspectNext(true);
				end);
				return; --to INSPECT_ACHIEVEMENT_READY
			else
				logRecord(_ProcessingID, name, "state", STATE["FAILED"]);
			end
		end
		inspectNext(true);
	else
		local sessionID = _ProcessingID;
		_ProcessingID = nil;
		wipe(_ProcessingList);
		InspectorFrame:SetSession(sessionID);
	end
end

local function newSession(id, unitPrefix)
	if not id or _ProcessingID then return end

	local sessionID;
	if DB.mergeRecord then
		for index, record in ipairs(_RecordTable) do
			if record.id == id then
				sessionID = index;
				break;
			end
		end
	end

	if not sessionID then
		sessionID = #_RecordTable + 1;
		local _, name, _, _, _, _, _, _, flags = GetAchievementInfo(id);
		_RecordTable[sessionID] = {};
		_RecordTable[sessionID].id = id;
		_RecordTable[sessionID].name = name or "unknown achievement";
		_RecordTable[sessionID].flag = flags or 0;
		if bit.band(_RecordTable[sessionID].flag, FLAG_STATISTIC) > 0 then
			_RecordTable[sessionID].metaData = META_STATISTIC;
		else
			_RecordTable[sessionID].metaData = META_ACHIEVEMENT;
		end
		_RecordTable[sessionID].data = {};

		InspectorFrame:UpdateSessionSelector();
	end
	_ProcessingID = sessionID;

	local unitData = _RecordTable[sessionID].data;
	if unitPrefix == "target" then
		local name = UnitName("target");
		unitData[name] = newUnitData("target");
		logRecord(_ProcessingID, name, "state", STATE["INITIAL"]);
		tinsert(_ProcessingList, "target");
	elseif unitPrefix == "party" then
		if unitPrefix == "party" then
			local name = UnitName("player");
			unitData[name] = newUnitData("player");
			logRecord(_ProcessingID, name, "state", STATE["INITIAL"]);
			tinsert(_ProcessingList, name);
		end
		for i = 1, GetNumPartyMembers() do
			local unitID = unitPrefix..i;
			local name = UnitName(unitID);
			unitData[name] = newUnitData(unitID);
			logRecord(_ProcessingID, name, "state", STATE["INITIAL"]);
			tinsert(_ProcessingList, name);
		end
	elseif unitPrefix == "raid" then
		for i = 1, GetNumRaidMembers() do
			local unitID = unitPrefix..i;
			local name = UnitName(unitID);
			unitData[name] = newUnitData(unitID);
			logRecord(_ProcessingID, name, "state", STATE["INITIAL"]);
			tinsert(_ProcessingList, name);
		end
	end
	--start inspect
	inspectNext();
end

local function inspectReady()
	if not _ProcessingID then return end

	SimpleTimer:CancelTimer(_name);
	local _, unitID = next(_ProcessingList);
	local name = UnitName(unitID);
	local record = _RecordTable[_ProcessingID];

	if bit.band(record.flag, FLAG_STATISTIC) > 0 then
		local info = GetComparisonStatistic(record.id) or "";
		if strlen(info) > 0 and info ~= "--" then
			logRecord(_ProcessingID, name, "state", STATE["DONE"], "info", info);
		else
			logRecord(_ProcessingID, name, "state", STATE["INCOMPLETED"], "info", info);
		end
	else
		local completed, month, day, year = GetAchievementComparisonInfo(record.id);
		completed = completed and STATE["COMPLETED"] or STATE["INCOMPLETED"];
		day = day and string.format(SHORTDATE, day, month, year) or nil;
		logRecord(_ProcessingID, name, "state", completed, "date", day);
	end

	ClearAchievementComparisonUnit();
	inspectNext(true);
end

---------------------------------------------------------------------------------------------btn function
local function refresh(sessionID)
	if _ProcessingID or not sessionID or not _RecordTable[sessionID] then return end
	_ProcessingID = sessionID;
	local data = _RecordTable[sessionID].data;
	for _, name in ipairs(InspectorFrame.entryIndex) do
		if data[name].state ~= STATE["COMPLETED"] then
			logRecord(sessionID, name, "state", STATE["INITIAL"]);
			tinsert(_ProcessingList, name);
		end
	end
	--start inspect
	inspectNext();
end

local function validateOutput(output)
	output = output and tostring(output) or "";
	if output:find(COINTEXTURETAG) then
		--textureString
		output = getShortCoinString(output);
	end
	return output;
end

local selected, incompleted = {}, {};
local function announce(sessionID, chatType, channel)
	local record = _RecordTable[sessionID];
	local valid, invalid, maxDataLen = 0, 0, 0;
	for _, name in ipairs(InspectorFrame.entryIndex) do
		if record.data[name].state == STATE["COMPLETED"] then
			valid = valid + 1;
		elseif record.data[name].state == STATE["DONE"] then
			valid = valid + 1;
		elseif record.data[name].state == STATE["INCOMPLETED"] then
			tinsert(incompleted, name);
		else
			invalid = invalid + 1;
		end
		if InspectorFrame.entryFrames[name]:IsSelected() then
			local info = validateOutput(record.data[name].date or record.data[name].info);
			tinsert(selected, name);
			tinsert(selected, info);
			maxDataLen = max(maxDataLen, strlen(name) + strlen(info));
		end
	end

	local msg = format("%s : %s (%d/%d/%d)", _name, GetAchievementLink(record.id), valid, valid + #incompleted, valid + #incompleted + invalid);
	SendChatMessage(msg, chatType, nil, channel);

	if #selected > 0 then
		if maxDataLen > 255 then
			am(_name.." : Announce length over 255 chars, please report this :3");
		else
			for i = 1, #selected, 2 do
				local name, info = selected[i], selected[i + 1];
				SendChatMessage(format("%s %"..maxDataLen.."s", name, info), chatType, nil, channel);
			end
		end
	elseif #incompleted > 0 then
		msg = "Incomplete : "..table.concat(incompleted, ",");
		maxDataLen = strlen(msg);
		if maxDataLen > 255 then
			for i = 1, maxDataLen, 255 do
				SendChatMessage(msg:sub(i, min(i + 254, maxDataLen)), chatType, nil, channel);
			end
		else
			SendChatMessage(msg, chatType, nil, channel);
		end
	end
	wipe(selected);
	wipe(incompleted);
end

local function bootSelected()
	for name, entry in pairs(InspectorFrame.entryFrames) do
		if entry:IsSelected() then
			UninviteUnit(name);
		end
	end
end

local function garbege(sessionID)
	if not sessionID or not _RecordTable[sessionID] then return	end
	wipe(_RecordTable[sessionID]);
	tremove(_RecordTable, sessionID);
	if sessionID > 1 then
		InspectorFrame:SetSession(sessionID - 1);
	elseif sessionID <= #_RecordTable then
		InspectorFrame:SetSession(sessionID);
	else
		InspectorFrame:SetSession(nil);
	end
end

---------------------------------------------------------------------------------------------menu/hook
local info = {};
local dummy = function() end;
local function dewdropAchievementPopMenu(id)
	if not Dewdrop:IsOpen() then return end
	local _, name, _, _, _, _, _, _, _, icon = GetAchievementInfo(id);
	info[1] = {
		text = name,
		icon = icon,
		iconHeight = 16,
		iconWidth = 16,
		isTitle = true,
		tooltipFunc = dummy,
	};
	info[2] = {
		text = L["QUICKINSPECT"],
		hasArrow = true,
		subMenu = {},
		disabled = _ProcessingID,
		tooltipFunc = dummy,
	};
	info[3] = {
		text = L["OPENINSPECTOR"],
		func = function()
			InspectorFrame:Show();
		end,
		tooltipFunc = dummy,
		disabled = InspectorFrame:IsShown(),
		closeWhenClicked = true,
	};
	info[4] = {
		text = "",
		textHeight = 5,
		disabled = true,
	};
	info[5] = {
		text = L["OPTIONS"],
		hasArrow = true,
		subMenu = {},
		disabled = _ProcessingID,
		tooltipFunc = dummy,
	};
	--QUICKINSPECT sub
	info[2].subMenu[1] = {
		text = UnitName("target") or L["TARGET"],
		func = function()
			newSession(id, "target");
		end,
		tooltipFunc = dummy,
		disabled = not UnitIsPlayer("target") or not UnitCanCooperate("player", "target"),
		closeWhenClicked = true,
	};
	local partyNum = GetNumPartyMembers();
	info[2].subMenu[2] = {
		text = L["PARTY"],
		func = function()
			newSession(id, "party");
		end,
		tooltipFunc = dummy,
		disabled = partyNum == 0,
		closeWhenClicked = true,
	};
	local raidNum = GetNumRaidMembers();
	info[2].subMenu[3] = {
		text = L["RAID"],
		func = function()
			newSession(id, "raid");
		end,
		tooltipFunc = dummy,
		disabled = raidNum == 0,
		closeWhenClicked = true,
	};
	--OPTIONS sub
	info[5].subMenu[1] = {
		text = L["MERGERECORD"],
		func = function()
			DB.mergeRecord = not DB.mergeRecord;
		end,
		tooltipFunc = dummy,
		notCheckable = false,
		checked = DB.mergeRecord,
		closeWhenClicked = true,
	};
	info[5].subMenu[2] = {
		text = L["TIMEOUT"],
		tooltipFunc = dummy,
		hasArrow = true,
		hasSlider = true,
		sliderMin = 1,
		sliderMax = 10,
		sliderValue = DB.timeout,
		sliderStep = 0.1,
		sliderFunc = function(value)
			DB.timeout = value;
		end,
	};
	--summary
	Dewdrop:FeedTable(info);
	wipe(info);
end

local function dewdropAnnouncePopupMenu()
	if not Dewdrop:IsOpen() then return end
	tinsert(info, {
		text = L["ANNOUNCE"],
		func = dummy,
		tooltipFunc = dummy,
		isTitle = true,
		closeWhenClicked = true,
	});
	tinsert(info, {
		text = L["A_SAY"],
		func = function()
			announce(InspectorFrame.sessionID, "SAY");
		end,
		tooltipFunc = dummy,
		closeWhenClicked = true,
	});
	if UnitIsPlayer("target") and UnitCanCooperate("player", "target") then
		tinsert(info, {
			text = L["A_WHISPER"]..UnitName("target"),
			func = function()
				announce(InspectorFrame.sessionID, "WHISPER", UnitName("target"));
			end,
			tooltipFunc = dummy,
			closeWhenClicked = true,
		});
	end
	if GetGuildInfo("player") then
		tinsert(info, {
			text = L["A_GUILD"],
			func = function()
				announce(InspectorFrame.sessionID, "GUILD");
			end,
			tooltipFunc = dummy,
			closeWhenClicked = true,
		});
	end
	if GetNumPartyMembers() > 0 then
		tinsert(info, {
			text = L["A_PARTY"],
			func = function()
				announce(InspectorFrame.sessionID, "PARTY");
			end,
			tooltipFunc = dummy,
			closeWhenClicked = true,
		});
	end
	if GetNumRaidMembers() > 0 then
		tinsert(info, {
			text = L["A_RAID"],
			func = function()
				announce(InspectorFrame.sessionID, "RAID");
			end,
			tooltipFunc = dummy,
			closeWhenClicked = true,
		});
	end
	local channelList = {GetChannelList()};
	for i = 1, #channelList, 2 do
		local id, name = channelList[i], channelList[i + 1];
		tinsert(info, {
			text = format("[%s. %s]", id, name),
			func = function()
				announce(InspectorFrame.sessionID, "CHANNEL", id);
			end,
			tooltipFunc = dummy,
			closeWhenClicked = true,
		});
	end
	Dewdrop:FeedTable(info);
	wipe(info);
end

local menuTrigger = function(self, id)
	if not id or not GetAchievementInfo(id) then return end
	local cursorX, cursorY = GetCursorPosition();
	Dewdrop:Open(self, "cursorX", cursorX, "cursorY", cursorY, "children", function()
		dewdropAchievementPopMenu(id);
	end);
end

local function hookAchievemenUItMenu()
	for _, button in next, AchievementFrameAchievements.buttons do
		local OriginClickScript = button:GetScript("OnClick");
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		button:SetScript("OnClick", function(...)
			local self, btn, down = ...;
			if btn == "RightButton" then
				menuTrigger(self, button.id);
			else
				OriginClickScript(...);
			end
		end);
	end
end

local function hookStaticUIEntry()
	for _, button in next, AchievementFrameStatsContainer.buttons do
		local OriginClickScript = button:GetScript("OnClick");
		button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		button:SetScript("OnClick", function(...)
			local self, btn, down = ...;
			if btn == "RightButton" then
				menuTrigger(self, button.id);
			else
				OriginClickScript(...);
			end
		end);
	end
end

local function overwriteEventHandler()
	AchievementFrameComparison:SetScript("OnEvent", function(self, ...)
		if self:IsVisible() then
			AchievementFrameComparison_OnEvent(self, ...);
		end
	end);
end

---------------------------------------------------------------------------------------------InspectorFrame
local EntryHeight = 16;

InspectorFrame = MessGUI:GetContainerFrame(UIParent);
InspectorFrame:SetTitle(_name);
InspectorFrame:UnregisterAllEvents();
InspectorFrame:Hide();
InspectorFrame.sessionID = nil;
InspectorFrame.entryIndex = {};	-- index : name, valid after Sort
InspectorFrame.entryFrames = {};	-- name : entryFrame
InspectorFrame.entryCount = 0;

function InspectorFrame:init()
	self:SetWidth(400);
	self:SetHeight(AchievementFrame:GetHeight());
	self:SetPoint("TOPLEFT", AchievementFrame, "TOPRIGHT");
	self.funcBtnGroup = {};
	--record selector
	local btn_l = CreateFrame("Button", "$ParentBtnL", self);
	btn_l:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up");
	btn_l:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down");
	btn_l:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Disabled");
	btn_l:SetWidth(24);
	btn_l:SetHeight(24);
	btn_l:SetPoint("TOPLEFT", 20, -20);
	btn_l:SetScript("OnClick", function()
		self:SetSession(self.sessionID - 1);
		self:SetStatusText(format("%d/%d", self.sessionID, #_RecordTable));
	end);
	btn_l:SetScript("OnEnter", function()
		self:SetStatusText(format("%d/%d", self.sessionID, #_RecordTable));
	end);
	btn_l:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	self.previousButton = btn_l;
	local btn_r = CreateFrame("Button", "$ParentBtnR", self);
	btn_r:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up");
	btn_r:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down");
	btn_r:SetDisabledTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Disabled");
	btn_r:SetWidth(24);
	btn_r:SetHeight(24);
	btn_r:SetPoint("TOPRIGHT", -20, -20);
	btn_r:SetScript("OnClick", function()
		self:SetSession(self.sessionID + 1);
		self:SetStatusText(format("%d/%d", self.sessionID, #_RecordTable));
	end);
	btn_r:SetScript("OnEnter", function()
		self:SetStatusText(format("%d/%d", self.sessionID, #_RecordTable));
	end);
	btn_r:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	self.nextButton = btn_r;
	--Recordlabel
	local label_recordID = self:CreateFontString(nil, "ARTWORK", "AchievementDateFont");
  label_recordID:SetText("");
  label_recordID:SetJustifyH("LEFT");
  label_recordID:SetPoint("LEFT", btn_l, "RIGHT", 10, 0);
  label_recordID:SetTextColor(LIGHTYELLOW_FONT_COLOR.r, LIGHTYELLOW_FONT_COLOR.g, LIGHTYELLOW_FONT_COLOR.b);
  self.label_recordID = label_recordID;
  local label_recordName = self:CreateFontString(nil, "ARTWORK", "AchievementDateFont");
  label_recordName:SetText("");
  label_recordName:SetJustifyH("RIGHT");
  label_recordName:SetPoint("RIGHT", btn_r, "LEFT", -10, 0);
  label_recordName:SetTextColor(LIGHTYELLOW_FONT_COLOR.r, LIGHTYELLOW_FONT_COLOR.g, LIGHTYELLOW_FONT_COLOR.b);
  self.label_recordName = label_recordName;
	--scrollpane
	local scrollPane = MessGUI:GetScrollPane(self, EntryHeight);
	scrollPane:SetPoint("TOPLEFT", btn_l, "BOTTOMLEFT", 0, -30);
	scrollPane:SetPoint("BOTTOMRIGHT", -20, 50);
	self.scrollPane = scrollPane;
	--table header
	local tabHeader = MessGUI:GetHorizontalDistributeFrame(self);
	tabHeader:SetPoint("BOTTOMLEFT", scrollPane.viewport, "TOPLEFT", 0, 10);
	tabHeader:SetPoint("BOTTOMRIGHT", scrollPane.viewport, "TOPRIGHT", 0, 10);
	self.tableHeader = tabHeader;
	--seperator
	local seperator_1 = self:CreateTexture(nil, "BACKGROUND");
	seperator_1:SetHeight(8);
	seperator_1:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border");
	seperator_1:SetTexCoord(0.81, 0.94, 0.5, 1);
	seperator_1:SetPoint("BOTTOMLEFT", scrollPane, "TOPLEFT");
	seperator_1:SetPoint("BOTTOMRIGHT", scrollPane, "TOPRIGHT");
	local seperator_2 = self:CreateTexture(nil, "BACKGROUND");
	seperator_2:SetHeight(8);
	seperator_2:SetTexture("Interface\\Tooltips\\UI-Tooltip-Border");
	seperator_2:SetTexCoord(0.81, 0.94, 0.5, 1);
	seperator_2:SetPoint("TOPLEFT", scrollPane, "BOTTOMLEFT");
	seperator_2:SetPoint("TOPRIGHT", scrollPane, "BOTTOMRIGHT");
	--function
	local btn_reload = MessGUI:GetIconButton(self, "Interface\\Icons\\Achievement_Boss_CThun");
	btn_reload:SetPoint("BOTTOMLEFT", 20, 20);
	btn_reload:SetWidth(24);
	btn_reload:SetHeight(24);
	btn_reload:SetScript("OnClick", function()
		refresh(self.sessionID);
	end);
	btn_reload:SetScript("OnEnter", function()
		self:SetStatusText(L["REFRESH"]);
	end);
	btn_reload:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	tinsert(self.funcBtnGroup, btn_reload);

	local btn_boradcast = MessGUI:GetIconButton(self, "Interface\\Icons\\Achievement_Halloween_Smiley_01");
	btn_boradcast:SetPoint("LEFT", btn_reload, "RIGHT", 10, 0);
	btn_boradcast:SetWidth(24);
	btn_boradcast:SetHeight(24);
	btn_boradcast:SetScript("OnClick", function()
		local cursorX, cursorY = GetCursorPosition();
		Dewdrop:Open(btn_boradcast, "cursorX", cursorX, "cursorY", cursorY, "children", dewdropAnnouncePopupMenu);
	end);
	btn_boradcast:SetScript("OnEnter", function()
		self:SetStatusText(L["ANNOUNCE"]);
	end);
	btn_boradcast:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	tinsert(self.funcBtnGroup, btn_boradcast);

	local btn_boot = MessGUI:GetIconButton(self, "Interface\\Icons\\Ability_Kick");
	btn_boot:SetPoint("LEFT", btn_boradcast, "RIGHT", 10, 0);
	btn_boot:SetWidth(24);
	btn_boot:SetHeight(24);
	btn_boot:SetScript("OnClick", function()
		bootSelected();
	end);
	btn_boot:SetScript("OnEnter", function()
		self:SetStatusText(L["BOOT"]);
	end);
	btn_boot:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	tinsert(self.funcBtnGroup, btn_boot);

	local btn_grabege = MessGUI:GetIconButton(self, "Interface\\Icons\\Spell_ChargeNegative");
	btn_grabege:SetPoint("LEFT", btn_boot, "RIGHT", 10, 0);
	btn_grabege:SetWidth(24);
	btn_grabege:SetHeight(24);
	btn_grabege:SetScript("OnClick", function()
		garbege(self.sessionID);
	end);
	btn_grabege:SetScript("OnEnter", function()
		self:SetStatusText(L["REMOVE"]);
	end);
	btn_grabege:SetScript("OnLeave", function()
		self:SetStatusText(L[""]);
	end);
	tinsert(self.funcBtnGroup, btn_grabege);
	--status lable
	local label_status = self:CreateFontString(nil, "ARTWORK", "GameFontNormal");
  label_status:SetText("");
  label_status:SetJustifyV("BOTTOM");
  label_status:SetJustifyH("RIGHT");
  label_status:SetPoint("BOTTOMRIGHT", -20, 20);
  label_status:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);
	self.label_status = label_status;

	self:SetSession(nil);
end

function InspectorFrame:SetStatusText(text)
	self.label_status:SetText(text);
end

function InspectorFrame:EnableFuncBtns(enable)
	if enable then
		for _, btn in ipairs(self.funcBtnGroup) do
			btn:Enable();
		end
	else
		for _, btn in ipairs(self.funcBtnGroup) do
			btn:Disable();
		end
	end
end

function InspectorFrame:UpdateSessionSelector()
	local sessionID = self.sessionID;
	if not sessionID then
		self.previousButton:Disable();
		self.nextButton:Disable();
	else
		if sessionID > 1 then
			self.previousButton:Enable();
		else
			self.previousButton:Disable();
		end
		if sessionID< #_RecordTable then
			self.nextButton:Enable();
		else
			self.nextButton:Disable();
		end
	end
end

local entryPool = {};
local function GetEntry(parent)
	local entry = tremove(entryPool);
	if not entry then
		entry = CreateFrame("Button", nil, parent);
		entry.fields = {};	-- key : fieldFrame
		entry:SetHeight(EntryHeight);

		local highlight = entry:CreateTexture(nil, "BACKGROUND");
	  highlight:SetTexture("Interface\\QuestFrame\\UI-QuestLogTitleHighlight");
	  highlight:SetBlendMode("ADD");
	  highlight:SetVertexColor(0.196, 0.388, 0.8);
	  highlight:SetTexCoord(0.1, 0.9, 0, 1);
	  highlight:SetAllPoints(entry);
	  entry:SetHighlightTexture(highlight);

	  entry.ClearAll = function(self)
	  	for _, fieldFrame in pairs(self.fields) do
	  		fieldFrame:Hide();
	  	end
	  end
	  entry.SetFieldText = function(self, key, text)
	  	if not self.fields[key] then
	  		self.fields[key] = self:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall");
	  	end
	  	self.fields[key]:SetText(text);
	  	self.fields[key]:Show();
	  end
	  entry.GetField = function(self, key)
	  	return self.fields[key];
	  end
	  entry.SetSelected = function(self, selected)
	  	self.selected = selected;
	  	if selected then
	  		self:LockHighlight();
			else
				self:UnlockHighlight();
			end
	  end
	  entry.IsSelected = function(self)
	  	return self.selected;
	  end
	  entry:SetScript("OnClick", function(self)
			self:SetSelected(not self.selected);
		end);
	else
		entry:SetParent(parent);
	end

	entry:SetSelected(false);
	entry:Show();
  return entry;
end

local headerPool = {};
local function GetHeaderEntry(parent, id)
	local entry = tremove(headerPool);
	if not entry then
		entry = MessGUI:GetMixButton(parent);
		entry:SetScript("OnClick", function(self)
			if self.id == DB.sortType then
				DB.sortSubtype = not DB.sortSubtype;
			else
				DB.sortType = self.id;
			end
			InspectorFrame:Resort();
		end);
	else
		entry:SetParent(parent);
	end
	entry.id = id;
	entry:Show();
  return entry;
end

local function SetEntryFields(entry, metaData, unitData, header)
	for _, field in ipairs(metaData) do
		--value
		local value = unitData[field];
		entry:SetFieldText(field, value);
		--layout
		local relatAlign = header:GetEntry(field);
		local point, _, _, xOfs, yOfs = relatAlign:GetPoint();
		local f = entry:GetField(field);
		f:ClearAllPoints();
		f:SetPoint("TOP");
  	f:SetPoint("BOTTOM");
		f:SetPoint(point, relatAlign);
	end
end

function InspectorFrame:UpdateRecordEntry(sessionID, name)
	if not self.sessionID then
		self:SetSession(sessionID);
	end
	if self.sessionID ~= sessionID then return end
	if name then
		local metaData = _RecordTable[sessionID].metaData;
		local unitData = _RecordTable[sessionID].data[name];	--assume data always prepared.
		if not self.entryFrames[name] then
			local entry = GetEntry(self.scrollPane.content);
			entry:ClearAllPoints();
			entry:SetPoint("LEFT");
			entry:SetPoint("RIGHT");
			entry:SetPoint("TOP", 0, -self.entryCount * EntryHeight);

			self.entryFrames[name] = entry;
			self.entryCount = self.entryCount + 1;
			self.scrollPane:SetContentHeight(self.entryCount * EntryHeight);
		end
		SetEntryFields(self.entryFrames[name], metaData, unitData, self.tableHeader);
	end
end

function InspectorFrame:SetSession(sessionID)
	local header = self.tableHeader;
	--release
	for _, entry in pairs(header:GetEntries()) do
		tinsert(headerPool, entry);
	end
	header:ClearAll();
	for _, entry in pairs(self.entryFrames) do
		entry:ClearAll();
		entry:Hide();
		tinsert(entryPool, entry);
	end
	wipe(self.entryFrames);
	self.entryCount = 0;
	self:EnableFuncBtns(false);
	--rebuild
	if not sessionID then
		self.label_recordID:SetText("");
		self.label_recordName:SetText("");
	else
		if not self:IsShown() then
			self:Show();
		end
		local record = _RecordTable[sessionID];
		self.label_recordID:SetText("ID:"..record.id);
		self.label_recordName:SetText(record.name);
		local metaData = record.metaData;
		for index, field in ipairs(metaData) do
			local headEntry = GetHeaderEntry(header, index);
			headEntry:SetText(field, "GameFontWhite");
			header:AddEntry(field, headEntry);
		end
		local data = _RecordTable[sessionID].data;
		for name, unitData in pairs(data) do
			local entry = GetEntry(self.scrollPane.content);
			entry:ClearAllPoints();
			entry:SetPoint("LEFT");
			entry:SetPoint("RIGHT");
			entry:SetPoint("TOP", 0, -self.entryCount * EntryHeight);

			SetEntryFields(entry, metaData, data[name], header);

			self.entryFrames[name] = entry;
			self.entryCount = self.entryCount + 1;
		end
		self.scrollPane:SetContentHeight(self.entryCount * EntryHeight);
		self:Sort(sessionID);
	end
	self.sessionID = sessionID;
	self:UpdateSessionSelector();
end

local function getCompareValue(value)
	value = value and tostring(value) or "";
	if tonumber(value) then
		--number 
		return tonumber(value), true;
	elseif value:find(COINTEXTURETAG) then
		--CoinTextureString
		return getCoinAmount(value), true;
	else
		--string
		return value, false;
	end
end

function InspectorFrame:Sort(sessionID)
	if not sessionID or _ProcessingID == sessionID then return end
	local metaData = _RecordTable[sessionID].metaData;
	local data = _RecordTable[sessionID].data;
	local field = metaData[DB.sortType];
	--header ui
	local header = self.tableHeader;
	for key, headEntry in pairs(header:GetEntries()) do
		if key == field then
			if DB.sortSubtype then
				headEntry:SetIcon(HEADERICON["DOWN"], HEADERICON["COORD_DOWN"]);
			else
				headEntry:SetIcon(HEADERICON["UP"], HEADERICON["COORD_UP"]);
			end
		else
			headEntry:SetIcon(nil);
		end
	end
	--sort entries
	wipe(self.entryIndex);
	for name, _ in pairs(self.entryFrames) do
		tinsert(self.entryIndex, name);
	end
	if field then
		table.sort(self.entryIndex, function(nameA, nameB)
			local A, isNumA = getCompareValue(data[nameA][field]);
			local B, isNumB = getCompareValue(data[nameB][field]);
			if isNumA or isNumB then
				--number compare, enusure A, B both number
				A = isNumA and A or 0;
				B = isNumB and B or 0;
			end
			if DB.sortSubtype then
				return A > B;
			else
				return A < B;
			end
		end);
		for index, name in ipairs(self.entryIndex) do
			local entry = self.entryFrames[name];
			entry:ClearAllPoints();
			entry:SetPoint("LEFT");
			entry:SetPoint("RIGHT");
			entry:SetPoint("TOP", 0, (1 - index) * EntryHeight);
		end
	end
	self:EnableFuncBtns(true);	--entryIndex is ready.
end

function InspectorFrame:Resort()
	self:Sort(self.sessionID);
end
---------------------------------------------------------------------------------------------
local addon = CreateFrame("Frame", nil, UIParent);
addon:UnregisterAllEvents();
addon:Hide();

function addon:ADDON_LOADED(name)
	if IsAddOnLoaded("Blizzard_AchievementUI") then
		--event when you first open achievement frame
		InspectorFrame:init();
		hookAchievemenUItMenu();
		hookStaticUIEntry();
		overwriteEventHandler();
		addon.hookflag = 1;
	elseif name == _name then
		if not AchieveInspectDB or AchieveInspectDB.version ~= defaultDB.version then
			AchieveInspectDB = CopyTable(defaultDB);
		end
		DB = AchieveInspectDB;
		addon.loadflag = 1;
		self:RegisterEvent("PLAYER_LOGIN");
	end
	if addon.hookflag and addon.loadflag then
		self:UnregisterEvent("ADDON_LOADED");
	end
end

function addon:PLAYER_LOGIN()
	self:UnregisterEvent("PLAYER_LOGIN");
	self:RegisterEvent("INSPECT_ACHIEVEMENT_READY");
end

function addon:INSPECT_ACHIEVEMENT_READY()
	inspectReady();
end


addon:SetScript("OnEvent", function(self, event, ...)
	local f = self[event];
	if f then
		f(self,...);
	end
end);

addon:RegisterEvent("ADDON_LOADED");